package audiovideo;

/**
 * This thread is responsible for,
 * If the client has the floor: capturing and recording audio packets and images
 * If the client doesn't hold the floor: playing back received packets
 * 
 * @author Tresor Mvumbi
 */
import java.awt.Image;
import java.awt.image.BufferedImage;
import java.awt.event.*;
import java.io.*;
import javax.media.Buffer;
import javax.media.control.FrameGrabbingControl;
import javax.media.format.VideoFormat;
import javax.media.util.BufferToImage;
import javax.sound.sampled.*;
import javax.swing.*;

import au.edu.jcu.v4l4j.FrameGrabber;
import au.edu.jcu.v4l4j.CaptureCallback;
import au.edu.jcu.v4l4j.V4L4JConstants;
import au.edu.jcu.v4l4j.VideoDevice;
import au.edu.jcu.v4l4j.VideoFrame;
import au.edu.jcu.v4l4j.exceptions.StateException;
import au.edu.jcu.v4l4j.exceptions.V4L4JException;
import java.awt.Color;
import javax.imageio.ImageIO;

public class CapturePlaybackThread extends Thread implements CaptureCallback { //

    // WIN VAR
    javax.media.CaptureDeviceInfo captureVideoDevice;
    VideoFormat captureVideoFormat;
    static public javax.media.Player player;
    private static String defaultVideoDeviceName = "vfw:Microsoft WDM Image Capture (Win32):0";//"Logitech USB Video Camera";
    private static String defaultVideoFormatString = "size=320x240, encoding=yuv, maxdatalength=153600"; //"size=176x144, encoding=yuv, maxdatalength=38016";
    public static BufferedImage image,  copyImage,  copyImage2;
    JLabel webCamViz = new JLabel();
    JFrame win;
    public static boolean videoInitialized = false;
//    public static javax.media.protocol.DataSource videoDataSource;
    
    
        ;
    
    // LINUX VAR
    private static int width = 320,  height = 240,  std = V4L4JConstants.STANDARD_WEBCAM,  channel = 0;
    private static String device = "/dev/video0";
    private VideoDevice videoDevice;
    private FrameGrabber frameGrabber;
    public VideoFrame frameVideo;
    public static int imagePacketNumber = 5;

    public CapturePlaybackThread(JFrame windows) {
        win = windows;
        try {
            if (!videoInitialized) {
                
                copyImage = ImageIO.read(new File("blank.png"));
                copyImage2 = ImageIO.read(new File("blank.png"));
            }

            // Initialization of video
            if (Framework.hasTheFloor && Framework.hasWebcam) {
                if (System.getProperty("os.name").toLowerCase().indexOf("win") >= 0) {
                    // Windows code
                    setCaptureDevice();
                } else {
                    // Linux code
                    initFrameGrabber();
                }
            }

            // Initialization of video player
            setVideoPlayer();

        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    @Override
    public void run() {
        try {
            while (!Framework.quit) {
                if (Framework.hasTheFloor && Framework.hasWebcam) { // If the client holds the floor, capture and record audio and video packets
                    if (System.getProperty("os.name").toLowerCase().indexOf("win") >= 0) {
//                        Thread.sleep(1000);
                        long time1 = System.currentTimeMillis();
                        captureImage();
                        long time2 = System.currentTimeMillis();
//                        System.out.println("Delay = "+(time2-time1));

                    } else {
                        initFrameGrabber();
                        initGUI();
                        frameGrabber.startCapture();
                    }
                } else { // If the client doesn't hold the floor, playback received packets
                    videoPlayBack();
                }

                if (!win.isVisible()) {
                    if (Framework.broadCastVisible) {
                        win.setVisible(true);
                    }
                } else {
                    if (!Framework.broadCastVisible) {
                        win.setVisible(false);
                    }
                }
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    private void captureImage() { // AND STREAM ...
        try {
            long time1 = System.currentTimeMillis();

            FrameGrabbingControl fgc = (FrameGrabbingControl) player.getControl("javax.media.control.FrameGrabbingControl");
//            Thread.sleep(2000);
            Buffer buf = fgc.grabFrame();
            // Convert it to an image
            BufferToImage btoi = new BufferToImage((VideoFormat) buf.getFormat());
            Image img = btoi.createImage(buf);

            image = (BufferedImage) img; // Image captured from the camera

            if (image != null) {
                imageDifferentiation(); // Image diff

                byte[] bufImage = getImageByteArray(image); // Generate the byte array corresonding

                if (Framework.hasWebcam) {// Send accross the network using HTTP
                    Broadcast broadcastWin = (Broadcast) win;
                    HttpBinaryTransfer.sendPacket(bufImage,
                            Framework.servletURL,
                            "video",
                            broadcastWin.user,
                            broadcastWin.meeting,
                            imagePacketNumber++);
                }
//                System.out.println("Image size : "+bufImage.length);
            }

            long time2 = System.currentTimeMillis();
            long delay = Framework.IMAGE_DURATION - (time2 - time1);
            if (delay > 0) {
                Thread.sleep(delay);
            }

        } catch (Exception e) {
            System.err.println("Capture image error");
            e.printStackTrace();
        }
    }

    private void imageDifferentiation() throws IOException {
//        BufferedImage bi = ImageIO.read(new File("pause.png"));
        // IMAGE DIFFERENTIATION
        if (imagePacketNumber % 10 != 0) {//Not a key frame, apply image diff
            Framework.copyBufferedImage(image, copyImage2);
            Framework.imageDiff(copyImage, image);
            Framework.copyBufferedImage(copyImage2, copyImage);
//            System.out.println("Diff calculated");
        } else {
            Framework.copyBufferedImage(image, copyImage);
//            System.out.println("no Diff calculated");
        }

//        System.out.println("Packet sent, number "+imagePacketNumber);
//        imagePacketNumber++;



    //  Convert image into a buffer 
//        if (new File("image.jpg").exists()) {
//            bi = ImageIO.read(new File("image.jpg"));
//            int w = image.getWidth();
//            int h = image.getHeight();
//            for (int i = 0; i < w; i++) {
//                for (int j = 0; j < h; j++) {
//                    int pxImg = image.getRGB(i, j);
//                    int pxBuff = bi.getRGB(i, j);
//                    Color colImg = new Color(pxImg);
//                    Color colBuff = new Color(pxBuff);
//                    Color colDiff = new Color(
//                            Math.abs(colBuff.getRed() - colImg.getRed()),
//                            Math.abs(colBuff.getGreen() - colImg.getGreen()),
//                            Math.abs(colBuff.getBlue() - colImg.getBlue()));
//                    //                            int pxDiff = pxImg - pxBuff; Math.abs
//                    bi.setRGB(i, j, colDiff.getRGB());
//                }
//            }

//            ImageIO.write(image,
//                    "jpg",
//                    new File("image.jpg"));
//        } else {
//            ImageIO.write(image,
//                    "jpg",
//                    new File("image.jpg"));
//        /*ImageIO.write(image,
//        "jpg",
//        new File("image123.jpg"));*/
//        }
    }

    public static byte[] getImageByteArray(BufferedImage bufImage) throws IOException {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        ImageIO.write(bufImage, "JPEG", baos);
        return baos.toByteArray();
    }

    private void videoPlayBack() {
//        try {
//            if (Framework.theBufImg != null) {
//                Thread.sleep(Framework.IMAGE_DURATION);
//                System.out.println("videoPlayBack");
//                ReceptionStream rc = (ReceptionStream) win;
////                JLabel theScreen = rc.screen;
////                webCamViz.setIcon(new ImageIcon(Framework.theBufImg));
//                rc.screen.setIcon(new ImageIcon(Framework.theBufImg));
//                win.getContentPane().repaint();
////                win.pack();
//                win.repaint();
//                System.out.println("Image displayed");
//            }
//        } catch (Exception e) {
//            e.printStackTrace();
//        }
    }

    private void setCaptureDevice() throws Exception {

        if (!videoInitialized) {
            videoInitialized = true;
            // Get list of all media devices
            java.util.Vector deviceListVector = javax.media.CaptureDeviceManager.getDeviceList(null);
            if (deviceListVector == null) {
                System.out.println("No media device detected (null object)");
                return;
            }
            if (deviceListVector.size() == 0) {
                System.out.println("No media device detected (0)");
                return;
            }

            // Get the format
            for (int x = 0; x < deviceListVector.size(); x++) {
                // display device name
                javax.media.CaptureDeviceInfo deviceInfo = (javax.media.CaptureDeviceInfo) deviceListVector.elementAt(x);
                String deviceInfoText = deviceInfo.getName();
                System.out.println(">>> Device info = " + deviceInfoText);

                javax.media.Format deviceFormat[] = deviceInfo.getFormats();
                for (int y = 0; y < deviceFormat.length; y++) {
                    // search for default video device
                    if (captureVideoDevice == null) {
                        if (deviceFormat[y] instanceof VideoFormat) {
                            System.out.println(">>> capture video device = " + deviceInfo.getName());
                            if (deviceInfo.getName().indexOf(defaultVideoDeviceName) >= 0) {
                                captureVideoDevice = deviceInfo;
                                System.out.println(">>> capture video device = " + deviceInfo.getName());
                            }
                        }
                    }

                    // search for default video format
                    if (captureVideoDevice == deviceInfo) {
                        if (captureVideoFormat == null) {
                            if (DeviceInfo.formatToString(deviceFormat[y]).indexOf(defaultVideoFormatString) >= 0) {
                                captureVideoFormat = (VideoFormat) deviceFormat[y];
                                System.out.println(">>> capture video format = " + DeviceInfo.formatToString(deviceFormat[y]));
                            }
                        }
                    }
                }
            }

            // Setup Video data source
            javax.media.MediaLocator videoMediaLocator = captureVideoDevice.getLocator();  // new MediaLocator("v4l2:///dev/video1");
            javax.media.protocol.DataSource videoDataSource = null;

            videoDataSource = javax.media.Manager.createDataSource(videoMediaLocator);
            DeviceInfo.setFormat(videoDataSource, captureVideoFormat);

            // Create a Player
            player = javax.media.Manager.createRealizedPlayer(videoDataSource);
            player.start();
        }

        if (Framework.hasTheFloor) {
//            win = new JFrame("Webcam");
//            win.setSize(320, 240);
//            win.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
//            win.getContentPane().setSize(320, 240);
//            win.getContentPane().add(player.getVisualComponent());
            Broadcast broadcastWin = (Broadcast) win;
            broadcastWin.screen.add(player.getVisualComponent());

//            win.pack();
//            win.setVisible(true);
        }
    }

    private void setVideoPlayer() throws Exception {
        if (false) {//!Framework.hasTheFloor
            System.out.println("Initialization of Webcam visualization");
//            win = new JFrame("Webcam visualization");
            win.setSize(320, 240);
            win.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
            win.getContentPane().setSize(320, 240);
            win.getContentPane().add(webCamViz);
            win.pack();
            win.setVisible(true);
        }
    }

    /**
     * Initialises the FrameGrabber object
     * @throws V4L4JException if any parameter if invalid
     */
    private void initFrameGrabber() throws V4L4JException {
        videoDevice = new VideoDevice(device);
        frameGrabber = videoDevice.getJPEGFrameGrabber(width, height, channel, std, 80);
        frameGrabber.setCaptureCallback(this);
        width = frameGrabber.getWidth();
        height = frameGrabber.getHeight();
        System.out.println("Starting capture at " + width + "x" + height);
    }

    /** 
     * Creates the UI components and initialises them
     */
    private void initGUI() {
//        win = new JFrame();
        webCamViz = new JLabel();
        win.getContentPane().add(webCamViz);
        win.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        //win.addWindowListener(this);
        win.setVisible(true);
        win.setSize(width, height);
    }

    /**
     * this method stops the capture and releases the frame grabber and video device
     */
    private void cleanupCapture() {
        try {
            frameGrabber.stopCapture();
        } catch (StateException ex) {
        // the frame grabber may be already stopped, so we just ignore
        // any exception and simply continue.
        }

        // release the frame grabber and video device
        videoDevice.releaseFrameGrabber();
        videoDevice.release();
    }

    @Override
    public void exceptionReceived(V4L4JException e) {
        // This method is called by v4l4j if an exception
        // occurs while waiting for a new frame to be ready.
        // The exception is available through e.getCause()
        e.printStackTrace();
    }

    @Override
    public void nextFrame(VideoFrame frame) {
        // This method is called when a new frame is ready.
        // Don't forget to recycle it when done dealing with the frame.

        // draw the new frame onto the JLabel
        image = frame.getBufferedImage();
        webCamViz.getGraphics().drawImage(image, 0, 0, width, height, null);

        frameVideo = frame;

        // recycle the frame
        frame.recycle();

//        try{Thread.sleep(1000);}catch(Exception e){}
    }
}
