package audiovideo;


/**
 * This thread is responsible for,
 * If the client has the floor: capturing and recording audio packets and images
 * If the client doesn't hold the floor: playing back received packets
 * 
 * @author Tresor Mvumbi
 */
import java.awt.Image;
import java.awt.Toolkit;
import java.awt.image.BufferedImage;
import java.io.*;
import javax.imageio.ImageIO;
import javax.media.Buffer;
import javax.media.control.FrameGrabbingControl;
import javax.media.format.VideoFormat;
import javax.media.util.BufferToImage;
import javax.sound.sampled.*;
import javax.swing.*;
import sun.audio.AudioStream;

public class CapturePlaybackThread1 extends Thread {

    AudioFormat audioFormat;
    TargetDataLine targetDataLine;
    javax.media.CaptureDeviceInfo captureVideoDevice;
    VideoFormat captureVideoFormat;
    public javax.media.Player player;
    JFrame win;
    private static String defaultVideoDeviceName = "vfw:Microsoft WDM Image Capture (Win32):0";//"Logitech USB Video Camera";
    private static String defaultVideoFormatString = "size=320x240, encoding=yuv, maxdatalength=153600"; //"size=176x144, encoding=yuv, maxdatalength=38016";
    public BufferedImage image;
    JLabel webCamViz = new JLabel();

    public CapturePlaybackThread1() {
        try {

            // Initialization of Audio for capture
            /*audioFormat = getAudioFormat();
            DataLine.Info dataLineInfo =
            new DataLine.Info(
            TargetDataLine.class,
            audioFormat);
            targetDataLine = (TargetDataLine) AudioSystem.getLine(dataLineInfo);*/

            // Initialization of video
            if (Framework.hasTheFloor && Framework.hasWebcam) {
                setCaptureDevice();
            }

            // Initialization of video player
            setVideoPlayer();

        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    @Override
    public void run() {
        try {
            while (!Framework.quit) {
                if (Framework.hasTheFloor && Framework.hasWebcam) { // If the client holds the floor, capture and record audio and video packets
//                    captureAudio(); // audio capture
                    // image capture
//                    System.out.println("I do have the floor");
                    captureImage();
                } else { // If the client doesn't hold the floor, playback received packets
//                    audioPlayBack();// audio playback

                    // video playback
//                    System.out.println("Loop1");
//                    System.out.println("I don't have the floor");
                    videoPlayBack();
//                    System.out.println("Loop");
                }
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    private void captureAudio() throws Exception {
        CaptureThread cap = new CaptureThread(audioFormat, targetDataLine);
        cap.fileName = "sent\\" + Framework.formatSequenceNumber(Framework.seqNumber) + "_" + Framework.clientId + "_" + Framework.meetingId + ".au";
        cap.start();
        Thread.sleep(Framework.SOUND_PACKET_DURATION);
        cap.stopCapture();

        Framework.seqNumber++;
    }

    private void captureImage() {
        try {
            FrameGrabbingControl fgc = (FrameGrabbingControl) player.getControl("javax.media.control.FrameGrabbingControl");
            Buffer buf = fgc.grabFrame();

            // Convert it to an image
            BufferToImage btoi = new BufferToImage((VideoFormat) buf.getFormat());
            Image img = btoi.createImage(buf);

            image = (BufferedImage) img;

            Thread.sleep(Framework.IMAGE_DURATION);

//            if (image != null) {
//                File outputfile = new File("image.jpg");
//                ImageIO.write(image, "jpg", outputfile);
//            }

        } catch (Exception e) {
            System.err.println("Capture image error");
            e.printStackTrace();
        }
    }

    private void videoPlayBack() {
        try {
//            File f = new File("received\\image.jpg");
//            if (!f.exists()) return;
//            
//            if (f.length() == 0) return;
//            
//            BufferedImage img = null;
//            img = ImageIO.read(f);
            if (Framework.theBufImg != null) {
                Thread.sleep(Framework.IMAGE_DURATION);
                System.out.println("videoPlayBack");
                webCamViz.setIcon(new ImageIcon(Framework.theBufImg));
                win.getContentPane().repaint();
                win.pack();
                win.repaint();
                System.out.println("Image displayed");
            }
        } catch (Exception e) {
            e.printStackTrace();
        //System.err.println("videoPlayBack error");
        }
    }

    private void setCaptureDevice() throws Exception {
//        System.out.println("Set Capture Device");
        // Get list of all media devices
        java.util.Vector deviceListVector = javax.media.CaptureDeviceManager.getDeviceList(null);
        if (deviceListVector == null) {
            System.out.println("No media device detected (null object)");
            return;
        }
        if (deviceListVector.size() == 0) {
            System.out.println("No media device detected (0)");
            return;
        }

        // Get the format
        for (int x = 0; x < deviceListVector.size(); x++) {
            // display device name
            javax.media.CaptureDeviceInfo deviceInfo = (javax.media.CaptureDeviceInfo) deviceListVector.elementAt(x);
            String deviceInfoText = deviceInfo.getName();

            javax.media.Format deviceFormat[] = deviceInfo.getFormats();
            for (int y = 0; y < deviceFormat.length; y++) {
                // search for default video device
                if (captureVideoDevice == null) {
                    if (deviceFormat[y] instanceof VideoFormat) {
                        if (deviceInfo.getName().indexOf(defaultVideoDeviceName) >= 0) {
                            captureVideoDevice = deviceInfo;
                            System.out.println(">>> capture video device = " + deviceInfo.getName());
                        }
                    }
                }

                // search for default video format
                if (captureVideoDevice == deviceInfo) {
                    if (captureVideoFormat == null) {
                        if (DeviceInfo.formatToString(deviceFormat[y]).indexOf(defaultVideoFormatString) >= 0) {
                            captureVideoFormat = (VideoFormat) deviceFormat[y];
                            System.out.println(">>> capture video format = " + DeviceInfo.formatToString(deviceFormat[y]));
                        }
                    }
                }
            }
        }

        // Setup Video data source
        javax.media.MediaLocator videoMediaLocator = captureVideoDevice.getLocator();
        javax.media.protocol.DataSource videoDataSource = null;

        videoDataSource = javax.media.Manager.createDataSource(videoMediaLocator);
        DeviceInfo.setFormat(videoDataSource, captureVideoFormat);

        // Create a Player
        player = javax.media.Manager.createRealizedPlayer(videoDataSource);
        player.start();

        if (Framework.hasTheFloor) {
            win = new JFrame("Webcam");
            win.setSize(320, 240);
            win.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
            win.getContentPane().setSize(320, 240);
            win.getContentPane().add(player.getVisualComponent());
            win.pack();
            win.setVisible(true);
        }


    }

    private void setVideoPlayer() throws Exception {
        if (!Framework.hasTheFloor) {
            System.out.println("Initialization of Webcam visualization");
            win = new JFrame("Webcam visualization");
            win.setSize(320, 240);
            win.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
            win.getContentPane().setSize(320, 240);
            win.getContentPane().add(webCamViz);
            win.pack();
            win.setVisible(true);
        }
    }

    private void audioPlayBack() throws Exception {
        File files[] = Framework.getFileListSorted(new File("received"));
        System.out.println(files.length);

        if (files.length < 3) {
            return;
        } // At least 2 files must be buffered before starting playback

        for (int i = 0; i < files.length; i++) {
            playAudioFile("received\\" + files[i].getName());
        }


    // possible optimization here, find a way to detect if new file arrived on the buffer and read them directly
    }

    private void playAudioFile(String fileName) throws Exception {
        InputStream in = new FileInputStream(fileName);
        AudioStream as = new AudioStream(in);
        sun.audio.AudioPlayer.player.start(as);
        Thread.sleep(Framework.SOUND_PACKET_DURATION - 500);

        as.close();
        in.close();

        new DeleteFileThread(new File(fileName)).start();
    }

    private AudioFormat getAudioFormat() {
        float sampleRate = 8000.0F;
        //8000,11025,16000,22050,44100
        int sampleSizeInBits = 16;
        //8,16
        int channels = 1;
        //1,2
        boolean signed = true;
        //true,false
        boolean bigEndian = false;
        //true,false
        return new AudioFormat(sampleRate,
                sampleSizeInBits,
                channels,
                signed,
                bigEndian);
    }
}

class CaptureThread extends Thread {

    AudioFormat audioFormat;
    TargetDataLine targetDataLine;
    public String fileName = "";

    public CaptureThread(AudioFormat audioF, TargetDataLine targetDL) {
        audioFormat = audioF;
        targetDataLine = targetDL;
    }

    @Override
    public void run() {
//        System.out.println("Start of capture");

        AudioFileFormat.Type fileType = null;
        File audioFile = null;

        fileType = AudioFileFormat.Type.AU;
        audioFile = new File(fileName);

        try {

            targetDataLine.open(audioFormat);
            targetDataLine.start();
            AudioSystem.write(
                    new AudioInputStream(targetDataLine),
                    fileType,
                    audioFile);

        } catch (Exception e) {
            e.printStackTrace();
        }//end catch
    }//end run
    public void stopCapture() {
        targetDataLine.stop();
//        System.out.println("End of capture");
    }

    public void closeCapture() {
        targetDataLine.stop();
        targetDataLine.close();
    }
}

class DeleteFileThread extends Thread {

    File file;

    public DeleteFileThread(File f) {
        file = f;
    }

    @Override
    public void run() {
//        System.out.println(file.isDirectory());

    //  System.out.println(file.delete());
    }
}
